#!/usr/bin/env python

import sys
from math import sqrt

"""
Created by: Alex Laird
Purpose: Project #1 in Survey of Programming Languages class
Date Created: 2/4/09
Date Modified: 3/24/09
Filename: adlIBMCP.py
Version: 0.2
         New in this version: debugging using the "-debug" call from
         the command line (or anytime while running the program) and
         also allows the user to type "-help" at anytime for a list
         of helpful hints. You can also now type "-credits" to give
         credit where credit's due.

The following is an application that emulates an IBM 650, one of the
few truly base-10 based hardware computers ever made.

The program would read programs as commands in the form of
~# ### ### ### where ~ was a + or - and the number blocks were memory
locations.  The first ~# was the operation to be performed on the
given numbers.  The second ### was the memory location of the first
number, the third ### was the memory location of the second number,
and the fourth ### was the destination memory location, where the
answer would be stored usually.

The commands are as follows:

f = function
s = sign

f\s |    +          |         -
--------------------|------------
 0  |     move      | negate and move
 1  |       +       |         -
 2  |       x       |         /
 3  |    square     |   square root
 4  |      =        |        !=
 5  |      >=       |         <
 6  |   x(y) -> z   |     x -> y(z)
 7  | incr and test |     (unused)
 8  |     read      |       print
 9  |     stop      |     (unused)

The program is read in in three section; initial data values, program
instructions, and input data.  The first section contains initialization
for certain memory locations.  The second section is where the messages
are stored which are to manipulate the memory locations, and the third
section, if used, is used on a +8 read operation.
"""

"""
Exit conditions that are returned for the Traceback.  Essentially error
messages, since for the project we were not to output anything except
proper data.
"""
# the program executed and ran properly
EXIT_CLEAN = 0

# an invalid input or output file was specified in the shell arguments
EXIT_INVALID_FILE = 1
EXIT_INVALID_INPUT_FILE = 2
EXIT_INVALID_OUTPUT_FILE = 3

"""
The emulator has reached the end of the input file (or received a blank line
from the shell) before receiving a stop "+9000000000" condition or the end
of the program message "+9999999999". In this scenario, the program may have
(and probably did) ran correctly, so the output could be correct, but it will
exit with a 2.
"""
EXIT_END_MESSAGE_NEVER_MET = 4

# the emulator received a message that was not 11 characters in length
EXIT_IMPROPER_MESSAGE_LENGTH = 5

# the message contained symbols other than +, -, or numerical digits
EXIT_INVALID_MESSAGE_CHARACTERS = 6

# the ~# command prefix was not recognized
EXIT_INVALID_PREFIX = 7

# you can't divide by zero with todays computers, so you certainly couldn't
# divide by zero in 1954
EXIT_DIVIDE_BY_ZERO = 8

"""
The emulator ran into a rogue pointer that is pointing at a memory location
outside of the allocated memory space.  This happens occassionally and frequently
indicates that the application is simply done.  The this is not an EXIT_CLEAN,
it is very likely that the input is still correct.
"""
EXIT_INVALID_POINTER = 9

# the command line arguments specified on runtime were not in the proper form
EXIT_INVALID_ARGUMENTS = 10

# for debugging, an array of string representations for the numerical exit codes
exitCodes = ["Clean exit",
             "Invalid file",
             "Invalid input file",
             "Invalid output file",
             "End message \"+9000000000\" or \"+9999999999\" never met",
             "Improper message length; a proper message is 11 characters",
             "Invalid message character; a proper message contains only a\n           leading + or - followed by ten numerical values",
             "Invalid prefix; a proper message can be prefixed with +0,\n           +1, +2, +3, +4, +5, +6, +7, +8, +9, -0, -1, -2,\n           -3, -4, -5,-6, or -8",
             "Divide by zero",
             "Invalid pointer; the indexPointer was to a location\n           that was not within the bounds of memory",
             "Invalid arguments; when running the emulator, it should be\n           called from the shell using the form\n           emulator.py input.txt output.txt -debug,\n           where -debug is optional; if an input or output\n           file is not present, usethe ' character as a\n           placeholder"]

# stores the output when in debug mode
outputForDebug = None
# set to true if debug mode is active (displays extensive results and error messages)
# or is inactive (exits silenty when an error occurs, displays only final result)
debug = False
# the stream that the input is parsed from, either a file or the shell
inputStream = None
# the writing stream that formats the output to either a file or the shell
outputStream = None
# pointer for the current message
indexPointer = 0
# an array that is essentially the virtual memory for the emulator
memory = []
# an array that contains the list of all commands read from the input
messages = []

"""
This class portrays a message object. It contains a string representation
of it self method simply for testing purposes.  The three parts of a message
are:

prefix: ~#
The prefix is a + or - (represented by ~) followed by a number (#) and is stored
as a string.
location1: ###
The first memory location can be up to three digits in length, from 0 to 999, and
is stored as an int.
location2: ###
The second memory location can be up to three digits in length, from 0 to 999, and
is stored as an int.
destination: ###
The destination memory location can be up to three digits in length, from 0 to 999,
and is stored as an int.
"""
class Message:
    # the operand
    prefix = None
    # the first memory location
    location1 = 0
    # the second memory location
    location2 = 0
    # the destination memory location
    destination = 0

    # for conversion to a string
    def __repr__(self):
        message = ""

        message += self.prefix

        # iterate through to ensure each item has three digits ###
        for item in [self.location1, self.location2, self.destination]:
            if item < 100:
                message += "0"
            if item < 10:
                message += "0"

            message += str(item)

        # return the string in the format ~######### where ~ is either a + or -
        return message

"""
This method ensures that the emulator's memory and variables are all reset
and read for a new program to be read.
"""
def init():
    global memory, messages, indexPointer

    memory = []
    messages = []
    indexPointer = 0

"""
This method cleans the message, removing whitespace and comments as needed
and only returning the first 11 characters of it; anything passed that point
is unnecessary and, therefore, ignored.

If this method returns None, that means the line was empty (except for spaces
or comments) but was not the end of the file, therefore it should just be
ignored by the emulator.
"""
def cleanMessage(message):
    message = message.replace(" ", "")

    # ignore comments or blank lines
    if len(message) == 0 or message[0] == "#" or message[0:2] == "//":
        return None

    # only accept the first eleven characters
    return message[0:11]

"""
This method reads the next line from the input stream. If the next line of
the file is "" (indicating the end of the file) or a blank line is sent from
the shell, the emulator terminates gracefully.
"""
def readNextLine():
    global inputStream

    # read the next line from the input
    message = inputStream.readline().rstrip('\n')
    
    # end of file condition
    if message == "":
        return EXIT_END_MESSAGE_NEVER_MET
    
    # check to see if the input was "-help" or "-credits" or "-debug"
    if (inputStream == sys.stdin and (message == "-help" or message == "-credits")) or message == "-debug":
        return message
    
    # clean the message to be proper
    return cleanMessage(message)

"""
This method is executed when input is from the shell and the user types either
"-help" or "-credits" or "-debug" instead of a proper message.
"""
def output(message):
    global debug
    
    if message == "-help":
        print("\n----------------------------")
        print("::Launching the Emulator::")
        print("When running this emulator from the command line, the comman should")
        print("appear in the following form:")
        print("emulator.py input.txt output.txt -debug\n")
        print("\"-debug\" is always optional, and input and output can be received")
        print("from the shell instead by providing an apostrophe (') instead of the")
        print("filename, as shown in the following example:")
        print("emulator.py ' output.txt -debug\n")
        print("will run the emulator with input from the shell, output to")
        print("output.txt, and debug mode active.")
        print("\n::Emulator Commands::")
        print("f = function")
        print("s = sign")
        print("\nf\s |    +          |         -")
        print("--------------------|------------")
        print(" 0  |     move      | negate and move")
        print(" 1  |       +       |         -")
        print(" 2  |       x       |         /")
        print(" 3  |    square     |   square root")
        print(" 4  |      =        |        !=")
        print(" 5  |      >=       |         <")
        print(" 6  |   x(y) -> z   |     x -> y(z)")
        print(" 7  | incr and test |     (unused)")
        print(" 8  |     read      |       print")
        print(" 9  |     stop      |     (unused)")
        print("\n::Emulator Overview::")
        print("his program reads programs as commands in the form of ~# ### ### ###")
        print("where ~ is a + or - and the number blocks are memory locations.  The")
        print("first ~# is the operation that will be performed on the given numbers.  The")
        print("second ### is the memory location of the first number, the third ### is the'")
        print("memory location of the second number, and the fourth ### is the destination")
        print("memory location, where the answer will be stored usually.")
        print("The program is read in in three section; initial data values, program")
        print("instructions, and input data.  The first section contains initialization")
        print("for certain memory locations.  The second section is where the messages")
        print("are stored which are to manipulate the memory locations, and the third")
        print("section, if used, is used on a +8 read operation.")
        print("\nExample: -0007000009 will negate and move (-0) the value in memory")
        print("location 7 (007) to memory location 9 (009).  In a case such as this,")
        print("notice that memory location 1 (000) is unused. This happens on simple")
        print("commands such as this.")
        print("\nAt any point when running the emulator, you can turn debug mode on and")
        print("off by typing \"-debug\".  Debug mode is not interactive, so there are")
        print("currently no additional commands you can enter while in debug mode; debug")
        print("mode merely offers more insite into the current state of the program when")
        print("commands are entered and manipulated.")
        print("----------------------------\n")
    elif message == "-credits":
        print("\n----------------------------")
        print("Created by: Alex Laird")
        print("Purpose: Project #1 in Survey of Programming Languages class")
        print("Date Created: 2/4/09")
        print("Date Modified: 3/24/09")
        print("Filename: adlIBMCP.py")
        print("Version: 0.2")
        print("         New in this version: debugging using the \"-debug\"")
        print("         call from the command line arguments. Also allows the")
        print("         user to type \"-help\" at any time for a list of helpful")
        print("         hints. You can also now type \"-credits\" to display")
        print("         information regarding the code itself ... Which you")
        print("         clearly just did. Congratulations!")
        print("----------------------------\n")
    elif message == "-debug":
        debug = not debug
        if debug == True:
            print("\n----------------------------")
            print("Debug mode is now active!")
            print("----------------------------\n")
        else:
            print("\n----------------------------")
            print("Debug mode is no longer active!")
            print("----------------------------\n")

"""
This method loops through reading from the input stream until two exit commands,
"+9999999999", are reached.  Before the first exit command, messages received
from the input stream are stored in the allocated memory.  After the first exit
command but before the second, messages are stored in a messages list to be
executed in the run() method.

If a message is not equal to 11 characters in length after it has been cleaned,
and the message cannot be cast seemlessly to an int, it is deemed an improper
message.
"""
def read():
    global memory, messages, inputStream
    
    debugSection = 0
    section = 0

    if debug == True:
        print("--READ INPUT--")

    while section < 2:
        if inputStream == sys.stdin:
            print(">> ", end = "")
            sys.stdout.flush()
        
        message = readNextLine()

        if debug == True:
            if debugSection == 0:
                if section == 0:
                    print("-INITIAL DATA VALUES (MEMORY)-")
                elif section == 1:
                    print("-PROGRAM COMMANDS (MESSAGES)-")
                debugSection += 1
            
            # display debugging results
            print("Message: " + message)
            
        # there is not a message on the line; skip it
        if message == None:
            continue
        # the message is a special case from shell input
        elif message == "-help" or message == "-credits" or message == "-debug":
            if inputStream == sys.stdin or message == "-debug":
                output(message)
            continue
        # the stop message +9000000000 was never met
        elif message == EXIT_END_MESSAGE_NEVER_MET:
            return EXIT_END_MESSAGE_NEVER_MET
        elif len(message) != 11:
            return EXIT_IMPROPER_MESSAGE_LENGTH
            
        # increment to the next section if the end of the section message is present
        if message == "+9999999999":
            section += 1
            debugSection -= 1
            continue

        if section == 0:
            # if the pass is in the first section, add a memory block
            try:
                memory += [int(message)]
            except:
                return EXIT_INVALID_MESSAGE_CHARACTERS
        else:
            # if the pass is in the second section, add the message to the list to be manipulated
            newMessage = Message()
            newMessage.prefix = message[0:2]
            newMessage.location1 = int(message[2:5])
            newMessage.location2 = int(message[5:8])
            newMessage.destination = int(message[8:11])

            messages += [newMessage]

    return 0

"""
This method actually runs the program that was received through the read()
method. It loops through each command in the messages list and parses it,
performing the specified operation.  This method will dynamically allocate
memory as needed.
"""
def run():
    global outputForDebug, debug, outputStream, memory, messages, indexPointer

    if debug == True:
        print("\n--EXECUTING COMMAND MESSAGES--")

    # loop until a +9 stop command is reached
    while True:
        # the pointer is invalid
        if indexPointer >= len(messages):
            return EXIT_INVALID_POINTER

        # fetch the message at the pointer
        m = messages[indexPointer]

        if debug == True:
            # debugging output
            print("\nIndex Pointer: " + str(indexPointer))
            print("Message: " + str(m))

        # increment the index pointer
        indexPointer += 1

        # allocate virtual memory
        for item in [m.location1, m.location2, m.destination]:
            if item >= len(memory):
                for x in range(item - len(memory)):
                    memory += [0]

        # parse the messages
        if m.prefix == "+0":
            # move
            memory[m.destination] = memory[m.location1]
            if debug == True:
                print("Operation: Move value from memory[location1] to memory[destination]")
        elif m.prefix == "+1":
            # add
            memory[m.destination] = memory[m.location1] + memory[m.location2]
            if debug == True:
                print("Operation: Add value from memory[location1] to memory[location2] and store")
                print("           memory[destination]")
        elif m.prefix == "+2":
            # multiply
            memory[m.destination] = memory[m.location1] * memory[m.location2]
            if debug == True:
                print("Operation: Multiple value from memory[location1 to memory[location2] and")
                print("           store in memory[destination]")
        elif m.prefix == "+3":
            # square
            memory[m.destination] = memory[m.location1] * memory[m.location1]
            if debug == True:
                print("Operation: Square value from memory[location1 and store in")
                print("           memory[destination]; memory[location2] is unused")
        elif m.prefix == "+4":
            # equal
            if memory[m.location1] == memory[m.location2]:
                indexPointer = m.destination
            if debug == True:
                print("Operation: If memory[location1] and memory[location2] are equal, set")
                print("           indexPointer to value in memory[destination]")
        elif m.prefix == "+5":
            # greater or equal
            if memory[m.location1] >= memory[m.location2]:
                indexPointer = m.destination
            if debug == True:
                print("Operation: If memory[location1] is greater than or equal to memory[location2],")
                print("           set indexPointer to value in memory[destination]")
        elif m.prefix == "+6":
            # increment memory size if needed
            if m.location1 + memory[m.location2] >= len(memory):
                memory += [0 for x in range((m.location1 + memory[m.location2] + 1) - len(memory))]

            # x(y) -> z
            memory[m.destination] = memory[m.location1 + memory[m.location2]]
            if debug == True:
                print("Operation: x(y) -> z")
                print("           Add location1 to memory[location2] and store value in")
                print("           memory[destination]")
        elif m.prefix == "+7":
            # increment and test
            memory[m.location1] += 1

            if memory[m.location1] != memory[m.location2]:
                indexPointer = m.destination
            if debug == True:
                print("Operation: Increment the value in memory[location1, test to see if it is now")
                print("           not equal to memory[location2; if they are not equal, set")
                print("           indexPointer to the value in memory[destination]")
        elif m.prefix == "+8":
            # read
            line = readNextLine()
            if line == "-help" or line == "-credits" or line == "-debug":
                output(line)
                continue
            else:
                memory[m.destination] = int(line)
            if debug == True:
                print("Operation: Read the next value from the input and store in memory[destination]")
        elif m.prefix == "+9":
            # the end
            if debug == True:
                print("Operation: Final message")
            return EXIT_CLEAN
        elif m.prefix == "-0":
            # negate and move
            memory[m.destination] = memory[m.location1] * (-1)
            if debug == True:
                print("Operation: Negate and move value from memory[location1] to memory[destination]")
        elif m.prefix == "-1":
            # subtract
            memory[m.destination] = memory[m.location1] - memory[m.location2]
            if debug == True:
                print("Operation: Subtract value from memory[location1] to memory[location2 and store")
                print("           in memory[destination]")
        elif m.prefix == "-2":
            # divide
            if debug == True:
                print("Operation: Divie value from memory[location1] to memory[location2] and store in")
                print("           memory[destination]")
            if(m.location2 == 0):
                return EXIT_DIVIDE_BY_ZERO
            else:
                memory[m.destination] = int(memory[m.location1] / memory[m.location2])
        elif m.prefix == "-3":
            # square root
            memory[m.destination] = int(sqrt(memory[m.location1]))
            if debug == True:
                print("Operation: Take the square root of the value in memory[location1 and")
                print("           store in memory[destination]")
        elif m.prefix == "-4":
            # not equal
            if memory[m.location1] != memory[m.location2]:
                indexPointer = m.destination
            if debug == True:
                print("Operation: If memory[location1] is not equal to memory[location2], set")
                print("           indexPointer to value in memory[destination]")
        elif m.prefix == "-5":
            # less than
            if memory[m.location1] < memory[m.location2]:
                indexPointer = m.destination
            if debug == True:
                print("Operation: If memory[location1] is less than memory[location2], set")
                print("           indexPointer to value in memory[destination]")
        elif m.prefix == "-6":
            # increment memory size if needed
            if m.location2 + memory[m.destination] >= len(memory):
                memory += [0 for x in range((m.location2 + memory[m.destination] + 1) - len(memory))]

            # x -> y(z)
            memory[m.location2 + memory[m.destination]] = memory[m.location1]
            if debug == True:
                print("Operation: x -> y(z)")
                print("           Add location2 to memory[destination] and use that sum as the")
        elif m.prefix == "-8":
            # write
            if debug == True:
                print("Operation: Write value in memory[location1] to the output")
                outputForDebug = memory[m.location1]
            outputStream.write(str(memory[m.location1]) + "\n")
        else:
            # somehow, someway, an invalid prefix was given to the program
            if debug == True:
                print("Operation: The prefix is invalid")
            return EXIT_INVALID_PREFIX

        if debug == True:
            print("Prefix: " + m.prefix)
            print("Memory Location 1: " + str(m.location1))
            print("Memory Location 2: " + str(m.location2))
            print("Memory Destination: " + str(m.destination))
            if outputForDebug != None:
                print("Output Value: " + str(outputForDebug))
                outputForDebug = None

"""
This method executes the emulator, instantiaiting the input/output streams,
calling methods to initialize the emulator, read the program data, and run
the program data, then returning an integer exit code (specified in detail
above) as the program exits.
"""
def main(args):
    global debug, inputStream, outputStream, memory

    if len(args) == 1 or (len(args) == 2 and args[1] == "-debug"):
        # use the shell for input and output
        inputStream = sys.stdin
        outputStream = sys.stdout
    elif len(args) == 2:
        try:
            # use specified file for input and shell for output
            inputStream = open(args[1], "r")
            outputStream = sys.stdout
        except:
                return EXIT_INVALID_INPUT_FILE
    elif len(args) == 3 or len(args) == 4:
        if args[1] == "'" and args[2] != "'":
            try:
                if args[2] != "-debug":
                    # use specified file for output and shell for input
                    inputStream = sys.stdin
                    outputStream = open(args[2], "w")
                else:
                    debug = True
                    return EXIT_INVALID_ARGUMENTS
            except:
                return EXIT_INVALID_OUTPUT_FILE
        elif args[2] == "'" and args[1] != "'" and len(args) == 4:
            try:
                if args[1] != "-debug":
                    # use specified file for input and shell for output
                    inputStream = open(args[1], "r")
                    outputStream = sys.stdout
                else:
                    debug = True
                    return EXIT_INVALID_ARGUMENTS
            except:
                if args[len(args) - 1] == "-debug":
                    debug = True
                return EXIT_INVALID_INPUT_FILE
        elif args[1] == "'" and args[2] == "'":
            # use shell for input and output
            inputStream = sys.stdin
            outputStream = sys.stdout
        elif args[1] != "'" and args[2] != "'" and args[1] != "-debug" and args[2] != "-debug":
            try:
                # use specified files for input and output
                inputStream = open(args[1], "r")
                outputStream = open(args[2], "w")
            except:
                return EXIT_INVALID_FILE
        else:
            debug = True
            return EXIT_INVALID_ARGUMENTS

    print("::IBM 650 Emulator v0.2 by Alex Laird::")
    if inputStream == sys.stdin:
        print("Type \"-help\" for help, \"-credits\" for credits, or start entering")
        print("commands if you think you know what you're doing.")
        print("Enter \"-debug\" at any time to enable or disable debug mode\n")
    else:
        print("")
            
    if args[len(args) - 1] == "-debug":
        debug = True
        print("---Debug Mode is Active---")
        print("Input Stream: " + str(inputStream))
        print("Output Stream: " + str(outputStream))
        print("\n--EMULATION STARTED--\n")
        
    # initialize emulator
    init()
    temp = [None for x in range(1000)]

    # execute the program
    code = read()
    if code == 0:
        if debug == True:
            print("-ALL MEMORY AND COMMAND MESSAGES READ-")
            print("-INPUT VALUES ARE READ WHEN CALLED-\n")
        # initialize memory
        temp[0:len(memory)] = memory
        memory = temp
        code = run()
    else:
        return code

    # if reading and writing from and to files, close connections
    if inputStream != sys.stdin:
        inputStream.close()
    if outputStream != sys.stdout:
        outputStream.close()
        
    return code

"""
Check for the main method and run it, passing it the system arguments
from the shell.
"""
if __name__ == "__main__":
    exitCond = main(sys.argv)
    if debug == True:
        if exitCond == 0:
            print("\n-ALL MESSAGES EXECUTED-")
        else:
            print("\n-ERROR IN MESSAGE EXECUTION-")
        print("Exit Code: " + str(exitCond))
        print("           " + exitCodes[exitCond])
        if exitCond == 0:
            print("-EMULATION ENDED-")
        elif exitCond == 4 or exitCond == 9:
            print("-EMULATION HALTED: minor error-")
        else:
            print("-EMULATION HALTED: fatal error-")
    sys.exit(exitCond)
